//
// File:        GenerateRmiClient.java
// Package:     gov.llnl.babel.backend.rmi
// Revision:    @(#) $Revision: 4434 $
// Date:        $Date: 2005-03-17 09:05:29 -0800 (Thu, 17 Mar 2005) $
// Description: Generate a Rmi client for a set of sidl symbols
// 
// Copyright (c) 2000-2001, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.backend.rmi2;

import java.util.Set;
import java.util.Iterator;
import java.util.TreeSet;

import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.CodeGenerator;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.Extendable;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolTable;

/**
 * <p>
 * This class is responsible for generating the Rmi client-side code. It
 * will generate Rmi extension types for classes and interfaces, and it will
 * generate native Rmi for enumerated types.
 * </p>
 */
public class GenerateRmiClient implements CodeGenerator {

   /**
    * Create a Rmi client generator.
    */
   public GenerateRmiClient() {
   }

   /**
    * Generate Rmi client-side code for each <code>SymbolID</code> in the
    * set argument. This is the initial entry point for generating the
    * client-side code. This routine assumes that all necessary symbols are
    * available in the symbol table. This method generates the client-side
    * code as a set of files and directories.
    *
    * @param symbols  a set of <code>SymbolID</code> objects.
    * @exception gov.llnl.babel.backend.CodeGenerationException
    *     provide feedback when code generation fails for one reason or
    *     another.
    * @see gov.llnl.babel.symbols.SymbolID
    */
   public void generateCode(Set symbols)
      throws CodeGenerationException
   {
      SymbolTable table = SymbolTable.getInstance();
      for(Iterator s = symbols.iterator(); s.hasNext() ; ){
         SymbolID id = (SymbolID)s.next();
         Symbol   symbol = table.lookupSymbol(id);
         if (symbol != null){
            switch(symbol.getSymbolType()){
            case Symbol.CLASS:
              generateExtendable((Extendable)symbol);
              break;
            case Symbol.INTERFACE:
            case Symbol.ENUM:
            case Symbol.PACKAGE:
              break;
            }
         }
      }
   }

  /**
   * Generate a Rmi extension in C to wrap the extendable.
   *
   * @param extendable		the extendable to wrap in Rmi
   *				C extension type.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    all catch all exception for any kind of problem while generating
   *    the package file.
   */
  private void generateExtendable(Extendable extendable)
    throws CodeGenerationException
  {
		CxxStubHeader stubH = new CxxStubHeader(extendable);
    CxxStubSource stubS = new CxxStubSource(extendable);
    stubH.generateCode();
    stubS.generateCode();

    CxxSkelSource skel = new CxxSkelSource(extendable);
    skel.generateCode();

	// RmiStub creates Impl files which invokes proteus
    RmiStubHeader rStubH = new RmiStubHeader(extendable);
    RmiStubSource rStubS = new RmiStubSource(extendable);
		rStubH.generateCode();
		rStubS.generateCode();
  }

  public String getType()
  {
    return "stub";
  }

  public boolean getUserSymbolsOnly()
  {
    return true;
  }

  public Set getLanguages()
  {
    Set result = new TreeSet();
    //result.add("c++");
    //result.add("cxx");
    result.add("rmi");
    return result;
  }
}
